import pkg_resources
from click import pass_context

from indico_install.updraft import init_tracking

def init(location):
    @pass_context
    def wrapper(ctx):
        """
        Initialize cluster configuration
        """
        template = pkg_resources.resource_filename(location, "cluster.yaml")
        ctx.invoke(init_tracking, input_yaml=template)

    return wrapper



from click import prompt
from indico_install.utils import run_cmd
from indico_install.infra.input_utils import postgres_input, auth_with_gsutil


def aks_setup():
    """
    Make sure we have the necessary scripts and auth for installation
    """
    assert run_cmd("which az", silent=True), "Please install the Azure CLI"
    assert run_cmd("which gcloud", silent=True) and run_cmd(
        "which gsutil", silent=True
    ), "Please install the gcloud and gsutil CLI"
    assert "name" in run_cmd(
        "az account show | grep name", silent=True
    ), "Please log in using az login"
    authed = auth_with_gsutil(raise_missing_key=True)
    assert authed, "Unable to auth with gcloud service account"


def ask_for_infra_input(conf):
    cluster = conf.get("cluster", {})
    conf["cluster"] = cluster
    storage_acct_name = prompt(
        "What is the name of the storage account",
        type=str,
        default=cluster.get("storage_acct"),
    )
    storage_acct_key = prompt(
        "What is a storage account key",
        type=str,
        default=cluster.get("storage_acct_key"),
    )
    rwx_fs = prompt(
        "Indico data fileshare name",
        type=str,
        default=conf["clusterVolumes"]["rwx"]["azureFile"]["shareName"],
    )
    rox_fs = prompt(
        "Indico api models fileshare name",
        type=str,
        default=conf["clusterVolumes"]["rox"]["azureFile"]["shareName"],
    )
    default_namespace = prompt(
        "What is the name of your default indico namespace?",
        type=str,
        default=conf.get("cluster", {}).get("defaultNamespace"),
    )
    conf["cluster"].update(
        {"storage_acct": storage_acct_name, "storage_acct_key": storage_acct_key}
    )
    conf["clusterVolumes"]["rwx"]["azureFile"]["shareName"] = rwx_fs
    conf["clusterVolumes"]["rox"]["azureFile"]["shareName"] = rox_fs
    conf["defaultNamespace"] = default_namespace
    postgres_input(conf)

    return conf



#!/usr/bin/env python3
import click


from indico_install.infra.init import init
from indico_install.cluster_manager import ClusterManager
from .setup import aks_setup, ask_for_infra_input
from . import storage, cluster


@click.group("aks")
@click.pass_context
def aks(ctx):
    """
    Indico infrastructure setup and validation for Azure Kubernetes Service
    """
    ctx.ensure_object(dict)
    ctx.obj["TRACKER"] = ClusterManager()


aks.command("init")(init(__name__))


@aks.command("check")
@click.pass_context
def check(ctx):
    """
    Check the state of an existing cluster to validate
    that it meets certain requirements
    """
    aks_setup()
    conf = ctx.obj["TRACKER"].cluster_config
    ask_for_infra_input(conf)
    cluster.check(conf)
    storage.check(conf)
    ctx.obj["TRACKER"].save()


@aks.command("create")
@click.pass_context
def create(ctx, upload=False):
    """
    Configure your AKS installation
    """
    aks_setup()
    conf = ctx.obj["TRACKER"].cluster_config
    ask_for_infra_input(conf)
    cluster.create(conf)
    storage.create(conf)
    ctx.obj["TRACKER"].save()



import click
import json
from OpenSSL import crypto
from click import prompt, secho

from indico_install.utils import run_cmd
from indico_install.infra.input_utils import download_indicoapi_data, auth_with_gsutil
from indico_install.kube.svc.restart import restart
from indico_install.infra.single_node import single_node
from indico_install.infra.gke import gke
from indico_install.infra.aks import aks
from indico_install.cluster_manager import ClusterManager


# Here we try and include the EKS function
# It depends on boto3 which may not be installed
# in which case we include a stub instead
try:
    from indico_install.infra.eks import eks
except Exception:

    @click.command("eks")
    def eks():
        """Not available"""
        pass

try:
    default_namespace = ClusterManager().cluster_config["defaultNamespace"]
except:
    default_namespace = "default"

@click.group("infra")
@click.pass_context
def infra(ctx):
    """
    Indico infrastructure setup and validation. Supports EKS, GKE, and single node installations.

    EKS Note! must install package with "eks" extras for EKS
    """
    pass


def update_secret(secret_name, namespace=default_namespace):
    """
    Update an existing secret
    """
    if run_cmd(f"kubectl get secrets -n {namespace} | grep {secret_name}", silent=True):
        del_secret = prompt(
            f"Generate secret: '{secret_name}'? If 'yes' existing secret will be deleted",
            type=bool,
            default=False,
        )
        if del_secret:
            run_cmd(f"kubectl delete secret -n {namespace} {secret_name}")
            return True
        else:
            return False
    else:
        return True

def generate_cert(
    country="US",
    state="MA",
    locality="Boston",
    org="Indico Data Solutions, Inc",
    org_unit="Engineering",
    common_name="indico.io",
    expiration_days=3652,
):
    key = crypto.PKey()
    key.generate_key(crypto.TYPE_RSA, 4096)
    cert = crypto.X509()
    cert.get_subject().C = country
    cert.get_subject().ST = state
    cert.get_subject().L = locality
    cert.get_subject().O = org
    cert.get_subject().OU = org_unit
    cert.get_subject().CN = common_name

    cert.set_issuer(cert.get_subject())
    cert.gmtime_adj_notBefore(0)
    cert.gmtime_adj_notAfter(int(expiration_days) * 24 * 60 * 60)

    cert.set_pubkey(key)
    cert.sign(key, "sha512")

    return {
        "crt": crypto.dump_certificate(crypto.FILETYPE_PEM, cert).decode(),
        "key": crypto.dump_privatekey(crypto.FILETYPE_PEM, key).decode(),
    }


@infra.command("generate-saml-cert")
@click.pass_context
@click.argument("host", required=True, type=str)
def generate_saml_cert(ctx, host=None):
    """
    Generate key/certificate for SSO
    """
    if not update_secret("indico-sso-cert"):
        return
    crypto_output = generate_cert(common_name=host)
    secho("Generating key/certificate", fg="green")
    run_cmd(
        f"""kubectl create secret generic indico-sso-cert --from-literal=sp.key="{crypto_output.get('key')}" --from-literal=sp.crt="{crypto_output.get('crt')}" """,
        silent=True,
    )
    ctx.invoke(restart, services=["noct"], contains=False, wait="2m")

try:
    from onelogin.saml2.idp_metadata_parser import OneLogin_Saml2_IdPMetadataParser
except ModuleNotFoundError:
    @infra.command("generate-saml-settings")
    def generate_saml_cert():
        """Not available"""
        secho("The generate-saml-settings command is not available. Please install required packages (i.e., 'sudo apt-get install libxml2-dev libxmlsec1-dev')  and then 'pip3 install indico-install[saml]'", fg="red")
else:
    @infra.command("generate-saml-settings")
    @click.pass_context
    @click.argument("host", required=True, type=str)
    @click.argument("metadata_url", required=False, type=str)
    @click.option(
        "-f",
        "--custom-saml",
        help="Optional path to a saml settings.json file; if specified, will upload the given file as a noct secret",
    )
    @click.option(
        "-a",
        "--advanced-settings",
        help="Optional path to a saml advanced_settings.json file; if specified, will upload the given file as a noct secret",
    )
    def generate_saml(
        ctx, host=None, metadata_url=None, custom_saml=None, advanced_settings=None
    ):
        """
        Generate and inject SAML settings from Azure AD Saml2
        Before running, complete the following steps:
        1. Register the indico platform as an app registration in Azure AD
        2. Add https://<indico IPA hostname>/auth/users/saml/acs" to the web redirection url whitelist

        HOST is the URL of the indico platform
        METADATA_URL is the federation metadata URL

        Example usage:
        indico infra generate-saml app.indico.io https://login.microsoftonline.com/12346-78910-11121314/federationmetadata/2007-06/federationmetadata.xml

        Alternatively, inject a custom settings.json file:
        indico infra generate-saml -f custom-settings.json

        An 'advanced_settings.json' file can be added with:
        indico infra generate-saml -f custom-settings.json -a advanced_settings.json

        See here for more details: https://github.com/onelogin/python3-saml

        """
        settings_secret_name = "indico-sso-settings"
        if run_cmd(f"kubectl get secrets | grep {settings_secret_name}", silent=True):
            del_secret = prompt(
                f"Secret named '{settings_secret_name}' already exists; remove and create a new secret?",
                type=bool,
                default=False,
            )
            if del_secret:
                run_cmd(f"kubectl delete secret {settings_secret_name}")
            else:
                return
        # load advanced_settings json if it exists
        adv_settings_content = ""
        if advanced_settings:
            with open(advanced_settings) as settings_file:
                adv_settings = json.load(settings_file)
            adv_settings_content = json.dumps(adv_settings, indent=4)
        advanced_secret = (
            f"--from-literal=advanced_settings.json='{adv_settings_content}'"
            if adv_settings_content
            else ""
        )
        if custom_saml:
            with open(custom_saml) as settings_file:
                settings = json.load(settings_file)
            settings_content = json.dumps(settings, indent=4)
            run_cmd(
                f"""kubectl create secret generic {settings_secret_name} --from-literal=settings.json='{settings_content}' """
                + f"{advanced_secret}",
                silent=True,
            )
            ctx.invoke(restart, services=["noct"], contains=False, wait="2m")
            return

        # not using custom settings
        if not (host and metadata_url):
            click.secho(
                "Please provide a host and metadata url if not using a custom settings.json file",
                fg="red",
            )
            return
        idp_info = OneLogin_Saml2_IdPMetadataParser.parse_remote(metadata_url)
        jacs = {
            "url": "https://" + host + "/auth/users/saml/acs",
            "binding": "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST",
        }
        jsls = {
            "url": "https://" + host + "/auth/users/saml/sls",
            "binding": "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect",
        }
        jsp = {
            "entityId": "https://" + host + "/auth/users/saml/metadata",
            "assertionConsumerService": jacs,
            "singleLogoutService": jsls,
            "NameIDFormat": "urn:oasis:names:tc:SAML:1.1:nameid-format:emailAddress",
        }
        stg = {"strict": "true", "debug": "true", "sp": jsp, "idp": idp_info["idp"]}
        settings_content = json.dumps(stg, indent=4)
        run_cmd(
            f"""kubectl create secret generic {settings_secret_name} --from-literal=settings.json='{settings_content}' """
            + f"{advanced_secret}",
            silent=True,
        )
        ctx.invoke(restart, services=["noct"], contains=False, wait="2m")


@infra.command("download-apidata")
@click.pass_context
@click.argument("version", required=True, type=str)
@click.option(
    "--extract/--no-extract",
    default=True,
    show_default=True,
    help="Automatically extract the downloaded TAR",
)
def download_api_data(ctx, version, *, extract):
    """
    Download VERSION of API data TAR from google cloud to local --deployment-root.
    Un-tar the file into a directory of the same name, also in the deployment-root
    VERSION is something like "v7".

    Requires Authentication with GSutil to download the TAR, but will attempt to auth with an existing key if it exists.

    Will not download TAR if it already exists.
    Will not extract the TAR if the data directory already exists in --deployment-root
    """
    auth_with_gsutil()
    download_indicoapi_data(version=version, extract=extract)


for command_group in [single_node, gke, eks, aks]:
    infra.add_command(command_group)


@infra.command("generate-metrics-cert")
@click.pass_context
def generate_metrics_cert(ctx):
    """
    Generate key/certificate for external metrics usage
    """
    crypto_output = generate_cert()
    run_cmd("kubectl create ns custom-metrics > /dev/null 2>&1", silent=True)
    if not update_secret("cm-adapter-serving-certs", namespace="custom-metrics"):
        return
    secho("Generating key/certificate", fg="green")
    run_cmd(
        f"""kubectl create secret generic cm-adapter-serving-certs -n custom-metrics --from-literal=serving.key="{crypto_output.get('key')}" --from-literal=serving.crt="{crypto_output.get('crt')}" """,
        silent=True,
    )



import click
from indico_install.utils import run_cmd, find_names


@click.command("restart")
@click.pass_context
@click.argument("services", required=True, nargs=-1)
@click.option(
    "--wait",
    default="0",
    show_default=True,
    help="Amount of time to wait for the new pods to be ready. Ex: 3m, 4h",
)
@click.option(
    "--contains/--no-contains",
    default=True,
    show_default=True,
    help=(
        "Restart all services with names containing the <service> string. "
        "Only applicable if <service> is not 'all'"
    ),
)
@click.option(
    "--all",
    "all_services",
    is_flag=True,
    default=False,
    help=(
        "Update all services in the current namespace (--all). "
        "Requires <service> passed in to be 'all'"
    ),
)
def restart(ctx, services, all_services=False, contains=True, wait="0"):
    """
    Reroll a K8S cluster deployment or statefulset.

    ARGS:

        <SERVICES> grep string(s) of deployments and statefulsets to reroll

         - "all" reset all deployments and statefulsets with an
     inditype label of service or celerytask
    """
    for svc_type in ["deployment", "statefulset"]:
        wait_str = "--watch=false" if wait == "0" else f"--timeout={wait}"
        if services[0] == "all":
            label_match = (
                "" if all_services else "-l 'inditype in (service, celerytask)'"
            )
            _services = run_cmd(
                (
                    f"kubectl get {svc_type} {label_match} --no-headers "
                    "-o custom-columns=NAME:.metadata.name"
                ),
                silent=True,
            ).splitlines()
            for svc_to_update in _services:
                svc_to_update = f"{svc_type}/{svc_to_update}"
                click.secho(f"service {svc_to_update}")
                run_cmd(
                    f"kubectl rollout restart {svc_to_update} "
                    f"&& kubectl rollout status {svc_to_update} {wait_str}"
                )
        else:
            service_str = r"\|".join(services)
            _services = find_names(svc_type, service_str, contains)
            for svc_to_update in _services:
                svc_to_update = f"{svc_type}/{svc_to_update}"
                run_cmd(
                    f"kubectl rollout restart {svc_to_update} "
                    f"&& kubectl rollout status {svc_to_update} {wait_str}"
                )



import base64
import os
import pty
import subprocess
from functools import wraps
from indico_install.config import D_PATH

import click



def options_wrapper():
    def opt_wrapper_outer(f):
        @wraps(f)
        @click.option(
            "--yes/--no",
            "-y/-n",
            default=False,
            show_default=True,
            help="Accept changes",
        )
        def wrapped(
            *args,
            yes=False,
            **kwargs,
        ):
            # Backend
            return f(
                *args,
                yes=yes,
                **kwargs,
            )

        return wrapped

    return opt_wrapper_outer


def _read(fd):
    return os.read(fd, 10240)


def with_pty(cmd: list):
    pty.spawn(cmd, master_read=_read, stdin_read=_read)


def run_cmd(mycmd, envvars: dict = None, silent=False, tty=False):
    if envvars is None:
        envvars = dict(os.environ)
    if tty and not isinstance(mycmd, list):
        raise AssertionError("Command must be passed in as a list for TTY")
    if not silent:
        click.secho(f"Running command: {mycmd}", fg="blue")

    popen_kwargs = dict(env=envvars, shell=True, stdout=subprocess.PIPE)
    if tty:
        return with_pty(mycmd)
    else:
        output = subprocess.Popen(mycmd, **popen_kwargs)
        stdout, _ = output.communicate()
        if stdout is not None:
            stdout = stdout.decode("utf-8").strip()
            if not silent:
                print(stdout, "\n")
        return stdout


def region_to_gmt(region):
    offset = "+0:00"
    if region.upper().startswith("us-east"):
        offset = "-5:00"
    return f"GMT{offset}"


def get_value(value_dict, keys):
    if not keys:
        return value_dict
    elif len(keys) == 1:
        return value_dict.get(keys[0], None)
    else:
        return get_value(value_dict[keys[0]], keys[1:])


def find_values(value_dict, target_value, base_key=None):
    if base_key is None:
        base_key = []
    for key, value in value_dict.items():
        if isinstance(value, dict):
            yield from find_values(value, target_value, base_key=base_key + [key])
        elif isinstance(value, str) and value.strip().lower() == target_value:
            yield value_dict, key, base_key


def base64file(filename):
    with open(filename, "r") as f:
        return convertb64(f.read())


def convertb64(mystring):
    return base64.b64encode(mystring.encode("utf-8")).decode("utf-8")


def decodeb64(mystring):
    return base64.b64decode(mystring.encode("utf-8")).decode("utf-8")


def determine_resource(deployment):
    for resource in ("deployment", "statefulset"):
        result = run_cmd(
            f"kubectl get {resource} | tail -n +2 | awk '{{print $1}}'", silent=True
        )
        if deployment in result.split():
            return resource


def current_user(clean=False):
    try:
        user = next(
            (
                s
                for s in run_cmd(
                    "gcloud auth list 2>/dev/null", silent=True
                ).splitlines()
                if s.startswith("*")
            ),
            "- unknown",
        ).split()[1]
        if user == "unknown":
            user = "service_acct" if os.getenv("GOOGLE_SERVICE_ACCOUNT") else None
        elif clean:
            user = user.split("@", 1)[0].replace("gcr-", "", 1)
        return user
    except Exception:
        return None


def current_tag():
    branch = run_cmd(
        f"git -C {D_PATH} rev-parse --abbrev-ref HEAD 2>&1", silent=True
    )
    if branch.startswith("fatal:"):
        return "unknown"
    commit = run_cmd(f"git -C {D_PATH} rev-parse 2>&1", silent=True)
    return string_to_tag(f"{branch}_{commit}")


def find_gcs_key(quiet=False):
    # TODO: improve the gcr key finder
    key_file = run_cmd(
        f"ls {D_PATH}" + r" | grep -e '^gcr-.*\.json$'", silent=True
    )
    err = None
    if not key_file:
        err = f"No GCR key file found in {D_PATH}! A GCR key file looks like 'gcr-<name>.json'"
    elif len(key_file.split("/n")) > 1:
        err = "Too many GCR key files found. Please only have 1"

    if err:
        if not quiet:
            click.secho(err, fg="red")
        return None
    return key_file


def string_to_tag(string):
    """Clean string to be used as folder path"""
    return "".join(
        [c if c.isalnum() or c in ("-", "_") else "_" for c in (string or "")]
    )


def pretty_diff(diff: list, show_all=False):
    for k, status, old_val, new_val in diff:
        if status == "+":
            click.secho(f"{k}: {new_val}", fg="green")
        elif status == "-":
            click.secho(f"{k}: {old_val}", fg="red")
        elif status == "*":
            click.secho(f"{k}: {old_val} -> {new_val}", fg="yellow")
        elif show_all:
            click.echo(f"{k}: {new_val}")


def diff_dicts(old: dict, new: dict):
    """
    Diff a dictionary
    + : added
    - : removed
    * : changed

    Return a list of tuples (key, status, old, new)
    """
    results = []
    for k in sorted(set(old).union(set(new))):
        old_val = old.get(k)
        new_val = new.get(k)
        status = None
        if old_val == new_val:
            pass
        elif old_val is None and new_val:
            status = "+"
        elif new_val is None and old_val:
            status = "-"
        elif new_val != old_val:
            status = "*"
        results.append((k, status, old_val, new_val))
    return results


def diff_files(f1, f2):
    diff_out = run_cmd(f"diff {f1} {f2}", silent=True)
    for line in diff_out.splitlines():
        if line.startswith("+"):
            click.secho(line, fg="green")
        elif line.startswith("-"):
            click.secho(line, fg="red")
        else:
            click.echo(line)
